import 'dart:async';

import 'package:jsonstreamreader/simpletoken.dart';

import 'jsontoken.dart';
import 'state/filestate.dart';
import 'streamer.dart';

//Produces a list of JsonTokens
class JsonResponse {
  int type;
  List<MapEntry<String, dynamic>> data;
  JsonResponse(this.type, this.data);
}

class JsonTokenizer {
  Streamer stream;
  JsonToken token;
  int identity;
  int _combine = 0;
  bool _called = false;
  JsonTokenizer(Streamer stream, {int combine = 0}) {
    this.stream = stream;
    this._combine = combine;
    this.identity = JsonToken.identity();
    JsonToken.initialize(identity);
    this.stream.setGroup(identity);
    this.token = new JsonToken(null, JsonToken.NULL,
        position: 0, group: this.identity, combine: this._combine);
  }
  JsonTokenizer.fromState(Streamer stream, int group, {int combine = 0}) {
    this.stream = stream;
    this._combine = combine;
    this.identity = group;
    JsonToken.initialize(identity);
    this.stream.setGroup(identity);
    if (JsonToken.queue[identity].length > 0) {
      this.token = JsonToken.queue[identity].values.last;
    }
  }
  Future<JsonResponse> tokenize() async {
    Completer<JsonResponse> completer = new Completer<JsonResponse>();
    this._called = false;
    this.stream.next().listen((List<SimpleToken> response) {
      this._called = true;
      try {
        response.forEach((SimpleToken rawToken) {
          if (rawToken.value.isNotEmpty) {
            JsonToken other = new JsonToken(rawToken.value, rawToken.type,
                group: this.identity, combine: this._combine);
            this.token = this.token.append(other);
          }
        });
      } catch (e) {
        new Timer(new Duration(microseconds: 0), () {
          if (!completer.isCompleted) {
            completer.completeError(e);
          }
        });
      }
    }, onDone: () {
      if (!this._called) {
        FileState.removeState(this.stream.file);
        if (!completer.isCompleted) {
          completer.complete(new JsonResponse(1, []));
        }
      } else {
        this.stream.setProgress();
        if (!completer.isCompleted) {
          completer
              .complete(new JsonResponse(0, JsonToken.batch[this.identity]));
        }
      }
    }, onError: (err) {
      if (!completer.isCompleted) {
        completer.completeError(err);
      }
    });
    return completer.future;
  }

  //remove completed items
  destroy() {
    if (this.token == null) {
      return;
    }
    List deletables = new List();
    List<int> keys = JsonToken.queue[this.identity].keys.toList();
    int len = keys.length;
    for (int m = 0; m < len; m++) {
      int key = keys[m];
      JsonToken item = JsonToken.queue[this.identity][key];
      if ((item.value == null || item.value == ',' || item.value == ':') &&
          item.position != this.token.position) {
        deletables.add(key);
      } else if (item.position != this.token.position) {
        if (item.parent != null) {
          if (item.parent.closed) {
            deletables.add(key);
          } else if (item.closed == true) {
            deletables.add(key);
          }
        } else if (item.closed == true) {
          deletables.add(key);
        }
      }
    }
    deletables.forEach((v) {
      JsonToken.queue[this.identity].remove(v);
    });
    //delete sendables
  }
}
