import 'dart:async';
import 'dart:convert';
import 'dart:io';
import 'remote/filestateprovider.dart';

///Normalizes a File and RemoteSource
class Normalized {
  dynamic _file;
  int _calls = 0;
  //last chunk fetch
  int _last = 0;
  bool _update = false;
  int _length = 0;
  int _end = 0;
  bool _lengthFound = false;
  bool _endResolved = false;
  String _runtime;
  Normalized(dynamic file) {
    this._file = file;
    this._runtime = this._file.runtimeType.toString();
  }

  ///temp path of downloaded file
  String get path {
    return this._file.path;
  }

  ///instance of File(dart.io)
  dynamic get file {
    return this._file;
  }

  ///identify range, where to start and stop reading file
  Normalized setBoundary(int start, int end) {
    this._end = end;
    this._last = start;
    return this;
  }

  void setLength(int length) {
    this._length = length;
    this._lengthFound = true;
  }

  ///delete a file
  Future<FileSystemEntity> delete() async {
    if (this._runtime != '_File') {
      return this._file.delete();
    } else {
      return this._file.getFile().delete();
    }
  }

  ///instance of File(dart.io)
  dynamic getFile() {
    return this._file;
  }

  Stream<List<int>> _next(int chunk) {
    int end = (this._last + chunk);
    int endProcess = 0;
    if (this._end > 0 && end >= this._end) {
      end = this._end;
      endProcess = 1;
    }
    if (!_lengthFound) {
      if (endProcess == 1) {
        _endResolved = true;
      }
      this._update = true;
      return lengthRead(this._last, end);
    } else {
      if (!_endResolved) {
        if (endProcess == 1) {
          _endResolved = true;
        }
        this._update = true;
        return this._file.openRead(this._last, end);
      } else {
        StreamController<List<int>> cc = new StreamController<List<int>>();
        new Timer(new Duration(microseconds: 0), () {
          cc.close();
        });
        return cc.stream;
      }
    }
  }

  Stream<List<int>> _fastNext(int chunk) {
    int end = (this._last + chunk);
    int endProcess = 0;
    if (this._end > 0 && end >= this._end) {
      end = this._end;
      endProcess = 1;
    }
    if (!_endResolved) {
      if (endProcess == 1) {
        _endResolved = true;
      }
      this._update = true;
      return this._file.openRead(this._last, end);
    } else {
      StreamController<List<int>> cc = new StreamController<List<int>>();
      new Timer(new Duration(microseconds: 0), () {
        cc.close();
      });
      return cc.stream;
    }
  }

  Stream<List<int>> lengthRead(int start, [int end = 0]) {
    StreamController<List<int>> cc = new StreamController<List<int>>();
    this._file.length().then((int l) {
      this._length = l;
      this._lengthFound = true;
      this._file.openRead(start, end).listen((List<int> line) {
        cc.add(line);
      }, onDone: () {
        cc.close();
      }, onError: (e) {
        cc.addError(e);
      });
    });
    return cc.stream;
  }

  ///return last number of bytes loaded
  int getLast() {
    return this._last;
  }

  ///get progress of read
  double getProgress() {
    if (this._end > 0) {
      return this._last >= this._end ? 100 : this._last / this._end;
    } else if (this._length == 0 &&
        this.isCompleted() &&
        this._last >= this.getCurrent() &&
        this.getCurrent() > 0) {
      return 1.0;
    } else if (this._length == 0 &&
        this.isCompleted() &&
        this.getCurrent() > 0) {
      return 0.9 - (0.3 / this._calls);
    } else if (this._length == 0) {
      return 0.4 - (0.4 / this._calls);
    } else {
      return this._last > this._length ? 100.0 : this._last / this._length;
    }
  }

  bool isCompleted() {
    return this._runtime == '_File' ? true : this._file.isCompleted();
  }

  int getCurrent() {
    return this._runtime == '_File' ? 0 : this._file.getCurrent();
  }

//get next range of bytes
  Stream<List<int>> next(int chunk, {bool findLength = true}) {
    this._calls++;
    this._last = !this._update ? this._last : this._last + chunk;
    return findLength ? this._next(chunk) : this._fastNext(chunk);
  }
}
