import 'dart:async';
import 'dart:convert';
import 'dart:io';

import 'package:jsonstreamreader/source/remote/fileawaite.dart';

import 'remote/resourcestate.dart';
import 'remote/filestateprovider.dart';

///Normalizes a File and RemoteSource
class Normalized {
  var _file;
  int _calls = 0;
  //last chunk fetch
  int _last = 0;
  int _length = 0;
  int _end = 0;
  bool _lengthFound = false;
  bool _reachedEndofStream = false;
  bool _endResolved = false;
  Normalized(var file) {
    this._file = file;
  }
  //temp path of downloaded file
  String get path {
    return this._file.path;
  }

  //instance of File(dart.io)
  dynamic get file {
    return this._file;
  }

  //identify range where to start and stop reading files
  Normalized setBoundary(int start, int end) {
    this._end = end;
    this._last = start;
    return this;
  }

//delete a file
  Future<FileSystemEntity> delete() async {
    String runtime = this._file.runtimeType.toString();
    if (runtime != '_File') {
      return this._file.delete();
    } else {
      return this._file.getFile().delete();
    }
  }

//instance of File(dart.io)
  dynamic getFile() {
    return this._file;
  }

  Stream<List<int>> _next(int chunk) {
    StreamController<List<int>> c = new StreamController<List<int>>();
    if (!this._lengthFound) {
      this._file.length().then((int l) {
        this._length = l;
        this._lengthFound = true;
        if (l == 0) {
          int end = (this._last + chunk);
          int endProcess = 0;
          if (this._end > 0 && end >= this._end) {
            end = this._end;
            endProcess = 1;
          }
          StreamSubscription<ResourceState> fstream;
          this._file.sp.pushState(end, FileStateProvider.CALLED);
          if (end < this._file.getCurrent() && !_endResolved) {
            this._file.getFile().openRead(this._last, end).listen(
                (List<int> line) {
              if (!c.isClosed) {
                c.add(line);
              } else {
                fstream.cancel();
              }
            }, onDone: () {
              this._last = end;
              if (endProcess > 0) {
                this._endResolved = true;
              }
              c.close();
            }, onError: (e) {
              c.addError(e);
            });
          } else if (end == this._file.getCurrent() && !_endResolved) {
            this._file.getFile().openRead(this._last, end).listen(
                (List<int> line) {
              if (!c.isClosed) {
                c.add(line);
              } else {
                fstream.cancel();
              }
            }, onDone: () {
              this._last = end;
              if (endProcess > 0) {
                this._endResolved = true;
              }
              c.close();
            }, onError: (e) {
              c.addError(e);
            });
          } else if (end > this._file.getCurrent() &&
              this._file.isCompleted() &&
              !this._reachedEndofStream &&
              !_endResolved) {
            this._reachedEndofStream = true;
            this._file.getFile().openRead(this._last).listen((List<int> line) {
              if (!c.isClosed) {
                c.add(line);
              } else {
                fstream.cancel();
              }
            }, onDone: () {
              this._last = this._file.getCurrent();
              if (endProcess > 0) {
                this._endResolved = true;
              }
              c.close();
            }, onError: (e) {
              c.addError(e);
            });
          } else if ((end > this._file.getCurrent() &&
                  this._file.isCompleted() &&
                  this._reachedEndofStream) ||
              this._endResolved) {
            new Timer(new Duration(microseconds: 0), () {
              c.add([124]);
              c.close();
            });
          } else if (!this._endResolved && !this._reachedEndofStream) {
            if (endProcess == 1) {
              this._endResolved = true;
              this._last = end;
            }
            new FileAwaite(end, endProcess, this._last, this._file.sp,
                (FileAwaiteResponse response) {
              if (response.state == FileAwaiteState.COMPLETED) {
                this._reachedEndofStream = true;
              }
              if (response.value == 0) {
                this._file.getFile().openRead(response.last).listen(
                    (List<int> line) {
                  if (!c.isClosed) {
                    c.add(line);
                  } else {
                    fstream.cancel();
                  }
                }, onDone: () {
                  this._last = this._file.getCurrent();
                  c.close();
                }, onError: (e) {
                  c.addError(e);
                });
              } else {
                this
                    ._file
                    .getFile()
                    .openRead(response.last, response.value)
                    .listen((List<int> line) {
                  if (!c.isClosed) {
                    c.add(line);
                  } else {
                    fstream.cancel();
                  }
                }, onDone: () {
                  this._last = response.value;
                  c.close();
                }, onError: (e) {
                  c.addError(e);
                });
              }
            });
          } else {
            new Timer(new Duration(microseconds: 0), () {
              if (!c.isClosed) {
                c.add([124]);
              }
              c.close();
            });
          }
        } else {
          int end = this._last + chunk;
          int endProcess = 0;
          if (this._end > 0 && end >= this._end) {
            end = this._end;
            endProcess = 1;
          }
          if (end > this._length && !_reachedEndofStream && !_endResolved) {
            _reachedEndofStream = true;
            this._file.openRead(this._last, this._length).listen(
                (List<int> line) {
              c.add(line);
            }, onDone: () {
              this._last = this._length;
              if (endProcess > 0) {
                this._endResolved = true;
              }
              c.close();
            }, onError: (e) {
              c.addError(e);
            });
          } else if (end < this._length && !_endResolved) {
            this._file.openRead(this._last, end).listen((List<int> line) {
              c.add(line);
            }, onDone: () {
              this._last = end;
              if (endProcess > 0) {
                this._endResolved = true;
              }
              c.close();
            }, onError: (e) {
              c.addError(e);
            });
          } else if (end == this._length && !_endResolved) {
            this._file.openRead(this._last, end).transform(utf8.decoder).listen(
                (List<int> line) {
              c.add(line);
            }, onDone: () {
              this._last = end;
              if (endProcess > 0) {
                this._endResolved = true;
              }
              c.close();
            }, onError: (e) {
              c.addError(e);
            });
          } else {
            new Timer(new Duration(microseconds: 0), () {
              c.add([124]);
              c.close();
            });
          }
        }
      }).catchError((error) {
        c.addError(error);
      });
    } else {
      if (this._length == 0) {
        int end = (this._last + chunk);
        int endProcess = 0;
        if (this._end > 0 && end >= this._end) {
          end = this._end;
          endProcess = 1;
        }
        StreamSubscription<ResourceState> fstream;
        this._file.sp.pushState(end, FileStateProvider.CALLED);
        if (end < this._file.getCurrent() && !_endResolved) {
          this._file.getFile().openRead(this._last, end).listen(
              (List<int> line) {
            if (!c.isClosed) {
              c.add(line);
            } else {
              fstream.cancel();
            }
          }, onDone: () {
            this._last = end;
            if (endProcess > 0) {
              this._endResolved = true;
            }
            c.close();
          }, onError: (e) {
            c.addError(e);
          });
        } else if (end == this._file.getCurrent() && !_endResolved) {
          this._file.getFile().openRead(this._last, end).listen(
              (List<int> line) {
            if (!c.isClosed) {
              c.add(line);
            } else {
              fstream.cancel();
            }
          }, onDone: () {
            this._last = end;
            if (endProcess > 0) {
              this._endResolved = true;
            }
            c.close();
          }, onError: (e) {
            c.addError(e);
          });
        } else if (end > this._file.getCurrent() &&
            this._file.isCompleted() &&
            !this._reachedEndofStream &&
            !_endResolved) {
          this._reachedEndofStream = true;
          this._file.getFile().openRead(this._last).listen((List<int> line) {
            if (!c.isClosed) {
              c.add(line);
            } else {
              fstream.cancel();
            }
          }, onDone: () {
            this._last = this._file.getCurrent();
            if (endProcess > 0) {
              this._endResolved = true;
            }
            c.close();
          }, onError: (e) {
            c.addError(e);
          });
        } else if ((end > this._file.getCurrent() &&
                this._file.isCompleted() &&
                this._reachedEndofStream) ||
            this._endResolved) {
          new Timer(new Duration(microseconds: 0), () {
            c.add([124]);
            c.close();
          });
        } else if (!this._endResolved && !this._reachedEndofStream) {
          if (endProcess == 1) {
            this._endResolved = true;
            this._last = end;
          }
          new FileAwaite(end, endProcess, this._last, this._file.sp,
              (FileAwaiteResponse response) {
            if (response.state == FileAwaiteState.COMPLETED) {
              this._reachedEndofStream = true;
            }
            if (response.value == 0) {
              this._file.getFile().openRead(response.last).listen(
                  (List<int> line) {
                if (!c.isClosed) {
                  c.add(line);
                } else {
                  fstream.cancel();
                }
              }, onDone: () {
                this._last = this._file.getCurrent();
                c.close();
              }, onError: (e) {
                c.addError(e);
              });
            } else {
              this
                  ._file
                  .getFile()
                  .openRead(response.last, response.value)
                  .listen((List<int> line) {
                if (!c.isClosed) {
                  c.add(line);
                } else {
                  fstream.cancel();
                }
              }, onDone: () {
                this._last = response.value;
                c.close();
              }, onError: (e) {
                c.addError(e);
              });
            }
          });
        } else {
          new Timer(new Duration(microseconds: 0), () {
            if (!c.isClosed) {
              c.add([124]);
            }
            c.close();
          });
        }
      } else {
        int end = this._last + chunk;
        int endProcess = 0;
        if (this._end > 0 && end >= this._end) {
          end = this._end;
          endProcess = 1;
        }
        if (end > this._length && !_reachedEndofStream && !_endResolved) {
          _reachedEndofStream = true;
          this._file.openRead(this._last, this._length).listen(
              (List<int> line) {
            c.add(line);
          }, onDone: () {
            this._last = this._length;
            if (endProcess > 0) {
              this._endResolved = true;
            }
            c.close();
          }, onError: (e) {
            c.addError(e);
          });
        } else if (end < this._length && !_endResolved) {
          this._file.openRead(this._last, end).listen((List<int> line) {
            c.add(line);
          }, onDone: () {
            this._last = end;
            if (endProcess > 0) {
              this._endResolved = true;
            }
            c.close();
          }, onError: (e) {
            c.addError(e);
          });
        } else if (end == this._length && !_endResolved) {
          this._file.openRead(this._last, end).transform(utf8.decoder).listen(
              (List<int> line) {
            c.add(line);
          }, onDone: () {
            this._last = end;
            if (endProcess > 0) {
              this._endResolved = true;
            }
            c.close();
          }, onError: (e) {
            c.addError(e);
          });
        } else {
          new Timer(new Duration(microseconds: 0), () {
            c.add([124]);
            c.close();
          });
        }
      }
    }
    return c.stream;
  }

  //return last number of bytes loaded
  int getLast() {
    return this._last;
  }

//get progress of read
  double getProgress() {
    if (this._end > 0) {
      return this._last / this._end;
    } else if (this._length == 0 && this._file.isCompleted()) {
      double p = this._last / this._file.getCurrent();
      if (p < 0.4) {
        p = 0.4 + 0.4 / ((1 - p) * 10);
      }
      return p;
    } else if (this._length == 0 &&
        this._file.isCompleted() &&
        this._last == this._file.getCurrent()) {
      return 1;
    } else if (this._length == 0) {
      return 0.4 - (0.4 / this._calls);
    } else if (this._length == 0 && this._file.isCompleted()) {
      return this._last / this._file.getCurrent();
    } else {
      return this._last / this._length;
    }
  }

//get next range of bytes
  Stream<List<int>> next(int chunk) {
    this._calls++;
    return this._next(chunk);
  }
}
