import 'dart:async';

//Responds when a range of bytes are request and loaded
class FileStateProvider {
  static const WAITING = 1;
  static const CALLED = 2;
  static const COMPLETED = 3;
  int _written = 0;
  StreamController<Null> _done = new StreamController<Null>.broadcast();
  StreamController<int> _write = new StreamController<int>.broadcast();
  //length:[state]
  Map<int, List<int>> calls = new Map<int, List<int>>();
  bool _isClosed = false;
  bool get isClosed {
    return _isClosed;
  }

  close() {
    _done.add(null);
    _done.close();
    _write.close();
    _isClosed = true;
  }

  Stream<int> until(int start, int unit) {
    StreamController<int> c = new StreamController<int>();
    StreamSubscription sb;
    bool found = false;
    if (!_write.isClosed) {
      sb = _write.stream.listen((int length) {
        found = true;
        if (length > start && (length < unit || unit == 0)) {
          if (!c.isClosed) {
            c.add(length);
          }
        } else if (length > start && (length == unit || unit == 0)) {
          if (!c.isClosed) {
            c.add(length);
            c.close();
            sb.cancel();
          }
        } else if (length > start && (length > unit || unit == 0)) {
          if (!c.isClosed) {
            //stop at unit not length
            c.add(unit == 0 ? _written : unit);
            c.close();
            sb.cancel();
          }
        }
      }, onDone: () {
        if (!found) {
          if (_written > start && (_written < unit || unit == 0)) {
            c.add(_written);
            c.close();
            sb.cancel();
          } else if (_written > start && (_written >= unit || unit == 0)) {
            c.add(unit == 0 ? _written : unit);
            c.close();
            sb.cancel();
          } else {
            c.close();
            sb.cancel();
          }
        } else if (!c.isClosed) {
          c.close();
          sb.cancel();
        } else {
          c.close();
          sb.cancel();
        }
      }, onError: (err) {
        if (!c.isClosed) {
          c.addError(err);
          sb.cancel();
        }
      });
    } else {
      if (_written > start && (_written < unit || unit == 0)) {
        c.add(_written);
        c.close();
      } else if (_written > start && (_written >= unit || unit == 0)) {
        c.add(unit == 0 ? _written : unit);
        c.close();
      } else {
        c.close();
      }
    }
    return c.stream;
  }

  //Describe current state of file read
  pushState(int length, int state) {
    if (state == CALLED) {
      if (calls[length] == null) {
        calls[length] = new List<int>();
        calls[length].add(state);
      } else {
        calls[length].add(state);
      }
    }
    if (state == WAITING) {
      _written = length;
      _write.add(length);
    }
  }

  //Identifes when all the entire file is downloaded
  Stream<Null> get done {
    return _done.stream;
  }
}
