import 'dart:async';
import 'dart:io';

import 'package:heavylist/heavylist.dart';
import 'package:jsonstreamreader/abstractreader.dart';

import 'complexexpression.dart';
import 'jsontoken.dart';
import 'jsontokenizer.dart';
import 'source/remote/remotesource.dart';
import 'state/globalreadermanager.dart';
import 'state/globalreaderstatemanager.dart';
import 'state/reader/streamreaderstate.dart';
import 'streamer.dart';
import 'streamitem.dart';

class StreamReader extends AbstractReader {
  JsonTokenizer tokenizer;
  StreamController<StreamItem> _trailing = new StreamController<StreamItem>();
  List<MapEntry<ComplexExpression, StreamController<StreamItem>>> _filters;
  StreamController<Null> _done = new StreamController<Null>.broadcast();
  StreamController<dynamic> _fail = new StreamController<dynamic>.broadcast();
  StreamController<double> _progress = new StreamController<double>.broadcast();
  StreamController<StreamReaderState> _onpause =
      new StreamController<StreamReaderState>.broadcast();
  Duration _delay;
  int _combine = 0;
  StreamReaderState _state;
  bool _paused = false;
  bool _running = false;
  HeavyList<MapEntry<String, dynamic>> _list;
  StreamController _parsing = new StreamController();
  bool _hault = false;
  String _reference = 'N';

  ///Create a new Reader from StreamReaderState
  StreamReader.fromState(StreamReaderState state) {
    GlobalReaderManager.addReader(this);
    _reference = 'S';
    this._delay = new Duration(microseconds: state.delay);
    this._filters =
        new List<MapEntry<ComplexExpression, StreamController<StreamItem>>>();
    this.tokenizer = new JsonTokenizer.fromState(
        new Streamer(new File('/empty')), state.group,
        combine: 1);
    _state = state;
    new Timer(new Duration(microseconds: 0), () {
      if (!_hault) {
        _stateFire();
      }
    });
  }
  void _stateFire() {
    this._parse(_state.parse).then((String d) {
      new Timer(_delay, () {
        this._partial(_state);
      });
    }).catchError((err) {
      if (_fail.hasListener) {
        _fail.add(err);
      }
      _onpause.close();
      _fail.close();
      //
      _done.close();
      _parsing.close();
      _progress.close();
      _trailing.close();
      if (this._filters != null) {
        if (this._filters.length > 0) {
          this._filters.forEach(
              (MapEntry<ComplexExpression, StreamController<StreamItem>> item) {
            item.value.close();
          });
        }
      }
    });
  }

  void _partial(StreamReaderState state) {
    Streamer stream;
    //recreate JsonTokens
    JsonToken.fromState(state.group, state.jsontokenstate);
    new Timer(new Duration(microseconds: 1), () {
      if (state.isRemote) {
        RemoteSource r = new RemoteSource.fromString(state.file);
        stream = new Streamer(r,
            start: state.start, end: state.end, chunksize: state.chunksize);
        new Timer(new Duration(microseconds: 0), () {
          r.download();
        });
      } else {
        File r = new File(state.file);
        stream = new Streamer(r,
            start: state.start, end: state.end, chunksize: state.chunksize);
      }
      this._combine = state.combine;
      this._delay = new Duration(microseconds: state.delay);
      this.tokenizer = new JsonTokenizer.fromState(stream, state.group,
          combine: this._combine);
      new Timer(new Duration(microseconds: 0), () {
        this._run();
      });
    });
  }

  ///@stream Steamer
  ///int delay how long should the reader wait before reading current and next sequence
  ///combine 1 converts idenified array values with the same parent into an object
  ///combine 2 uses the first common parent
  StreamReader(Streamer stream, {int delay = 0, combine = 0}) {
    this._combine = combine;
    this._delay = new Duration(microseconds: delay);
    this._filters =
        new List<MapEntry<ComplexExpression, StreamController<StreamItem>>>();
    this.tokenizer = new JsonTokenizer(stream, combine: this._combine);

    new Timer(new Duration(microseconds: 0), () {
      if (!_hault) {
        _fire();
      }
    });
  }
  void _fire() {
    new Timer(new Duration(microseconds: 0), () {
      this._run();
    });
  }

  _run() {
    this.tokenizer.tokenize().then((JsonResponse item) {
      if (item.type == 0) {
        this._parse(JsonToken.batch[this.tokenizer.identity]).then((d) {
          _run();
        }).catchError((e) {
          if (!_fail.isClosed) {
            if (_fail.hasListener) {
              _fail.add(e);
            }
            _onpause.close();
            _fail.close();
            //
            _done.close();
            _progress.close();
            _parsing.close();
            _trailing.close();
            this._filters.forEach(
                (MapEntry<ComplexExpression, StreamController<StreamItem>>
                    item) {
              item.value.close();
            });
          }
        });
      } else {
        if (_done.hasListener) {
          _done.add(null);
        }
        if (_progress.hasListener) {
          _progress.add(100.0);
        }
        _onpause.close();
        _done.close();
        //
        _fail.close();
        _progress.close();
        _trailing.close();
        _parsing.close();
        this._filters.forEach(
            (MapEntry<ComplexExpression, StreamController<StreamItem>> item) {
          item.value.close();
        });
      }
    }).catchError((err) {
      if (!_fail.isClosed) {
        if (_fail.hasListener) {
          _fail.add(err);
        }
        _onpause.close();
        _fail.close();
        //
        _done.close();
        _progress.close();
        _parsing.close();
        _trailing.close();
        this._filters.forEach(
            (MapEntry<ComplexExpression, StreamController<StreamItem>> item) {
          item.value.close();
        });
      }
    });
  }

  Future<String> _parse(List<MapEntry<String, dynamic>> parse) async {
    this.tokenizer.destroy();
    Completer<String> completer = new Completer<String>();
    _list = new HeavyList<MapEntry<String, dynamic>>(parse);
    _parsing.add(null);
    _running = true;
    _list.loop(new Duration(microseconds: 0),
        (List<MapEntry<String, dynamic>> origin) {
      _list.empty();
      double p = Streamer.progess(this.tokenizer.identity);
      if (_progress.hasListener) {
        _progress.add(p);
      }
      if (p >= 100) {
        _onpause.close();
        _progress.close();
        _trailing.close();
        this._filters.forEach(
            (MapEntry<ComplexExpression, StreamController<StreamItem>> item) {
          item.value.close();
        });
      }
      completer.complete("done");
    }, (MapEntry<String, dynamic> item, Function resume) {
      dynamic firstValue = item.value;
      dynamic firstKey = item.key;
      Iterable<String> ikeys = [];
      List<StreamController> functions = [];
      if (firstValue.runtimeType.toString() ==
          '_InternalLinkedHashMap<String, dynamic>') {
        ikeys = item.value.keys;
      }
      if (this._trailing.hasListener) {
        functions.add(this._trailing);
      }
      this
          ._filters
          .forEach((MapEntry<ComplexExpression, StreamController> filter) {
        if (filter.key.evaluate(ikeys, firstKey) && filter.value.hasListener) {
          functions.add(filter.value);
        }
      });
      if (functions.length > 0) {
        functions.forEach((StreamController s) {
          s.add(new StreamItem(firstValue, firstKey));
        });
        this._list.clearRegistered();
        resume();
      } else {
        this._list.clearRegistered();
        resume();
      }
    });
    return completer.future;
  }

  ///return all objects or values found
  Stream<StreamItem> get trailing {
    return _trailing.stream;
  }

  ///called when reader is complete
  Stream<Null> get done {
    return _done.stream;
  }

  ///called when reader fails
  Stream<dynamic> get fail {
    return _fail.stream;
  }

  ///double progess of streamreader
  Stream<double> get progress {
    return _progress.stream;
  }

  ///called everytime reader is paused
  Stream<StreamReaderState> get onpause {
    return _onpause.stream;
  }

  ///Prevent the reader from starting immediately
  void hault() {
    _hault = true;
  }

  //if the reader has hault,start now
  void begin() {
    if (_hault) {
      _hault = false;
      if (this._reference == 'N') {
        _fire();
      } else {
        _stateFire();
      }
    }
  }

  ///Filter allows users to filter object or values by JsonPath(eg $.items.0) or PathExpression(eg. /items/0/)
  Stream<StreamItem> filter(String expression) {
    this._filters.add(new MapEntry(
        new ComplexExpression(expression), new StreamController<StreamItem>()));
    return this._filters.last.value.stream;
  }

  ///remove filter at index
  void removeFilter(int index) {
    this._filters.removeAt(0);
  }

  ///Return the StreamReaderState was in upon pausing
  Future<StreamReaderState> pause() async {
    Completer<StreamReaderState> completer = new Completer<StreamReaderState>();
    if (!this._paused) {
      if (_running) {
        new Timer(new Duration(microseconds: 0), () {
          this._pause();
          completer.complete(this._state);
        });
      } else {
        StreamSubscription sb;
        sb = _parsing.stream.listen((onData) {
          this._pause();
          sb.cancel();
          if (!completer.isCompleted) {
            completer.complete(this._state);
          }
        });
      }
    }
    return completer.future;
  }

  void _pause() {
    this._paused = true;
    this._list.pause();
    this._state = new StreamReaderState(
        JsonToken.getState(this.tokenizer.identity),
        this.tokenizer.identity,
        this.tokenizer.stream.path,
        this.tokenizer.stream.start,
        this.tokenizer.stream.end,
        this._delay.inMicroseconds,
        this._combine,
        this.tokenizer.stream.chunksize,
        this._list.getList());
    if (_onpause.hasListener) {
      _onpause.add(this._state);
    }
    GlobalReaderStateManager.addState(this._state);
  }

  ///resume streamreaders
  void resume() {
    if (this._paused) {
      this._paused = false;
      this._list.resume();
      GlobalReaderStateManager.removeState(this._state);
    }
  }
}
